/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Dict
*  SCOPE :              Dict common services
*  FILE :               dict.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma implementation
#endif

/* IMPORTS *******************************************************************/

#ifndef __dict_HXX
#include "dict.hxx"
#endif

/*****************************************************************************/

/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> DictLeaf::clone()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return new(this->getCluster(), this->getClass()) DictLeaf(0) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<ColNode> DictLeaf::newNode()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return new(this->getCluster(), YOO_DICTNODE_CLID) DictNode(0) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
const yoo_ObjectX& DictLeaf::max()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (this->_card == 0) {
    return NULL_REF ;
  } else {
    return this->tab(this->_card - 1)->ref() ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
int
DictLeaf::indKey(const void* key,void* (*extract)(const void*),
		 int (*compare)(const void*, const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  void *atom, *currKey ;
  int flg, curr ;
  int inf = 0 ;
  int sup = _card - 1 ;

  if (_card == 0) return COL_TAIL ;

  // Dychotomic search into the leaf
  while (sup-inf > 1) {
    curr = (inf + sup) / 2 ;
    atom = this->tab(curr)->data() ;
    currKey = extract(atom) ;
    flg = compare(currKey, key) ;
    if (flg == 0) {
      return curr ;
    } else if (flg == 1) {
      sup = curr ;
    } else {
      inf = curr ;
    }
  }

  // If not found, test edges.
  atom = this->tab(inf)->data() ;  
  currKey = extract(atom) ;
  flg = compare(currKey, key) ;
  if (flg != 2) {
    return inf ;
  } else {
    atom = this->tab(sup)->data() ;
    currKey = extract(atom) ;
    flg = compare(currKey, key) ;
    if (flg != 2) {
      return sup ;
    } else {
      return COL_TAIL ;
    }
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictLeaf::yoo_refresh(void* addr)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  YOO_LOAD_TIME = TRUE ;
  new(addr) DictLeaf ;
  YOO_LOAD_TIME = FALSE ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> DictNode::clone()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return new(this->getCluster(), this->getClass()) DictNode(0) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
ColElem* DictNode::tab(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return (ColElem*) &(this->_tab[pos*sizeof(DictNodeElem)]) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictNode::sizeElem()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return sizeof(DictNodeElem ) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictNode::maxElem()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return DictNbInNode ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictNode::newChild(const void* atom, int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->ColNode::newChild(atom, pos) ;
  ((DictNodeElem*)this->tab(pos))->max = this->tab(pos)->ref()->max() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictNode::resetChild(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->ColNode::resetChild(pos) ;
  ((DictNodeElem*)this->tab(pos))->max = this->tab(pos)->ref()->max() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
const yoo_ObjectX& DictNode::max()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return ((DictNodeElem*)this->tab(_nbElem-1))->max ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int
DictNode::indKey(const void* key,void* (*extract)(const void*),
		 int (*compare)(const void*,const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  yoo_ObjectX tmp ;
  void* atom, *currKey ;
  int flg, curr ;
  int inf = 0 ;
  int sup = _nbElem - 1 ;

  if (_nbElem == 0) return COL_TAIL ;

  atom = (void*)&this->max() ;
  currKey = extract(atom) ;
  flg = compare(currKey, key) ;
  if (flg == 2) {
    return COL_TAIL ;
  }

  while (sup-inf > 1) {
    curr = (inf + sup) / 2 ;
    atom = &this->max(curr) ;
    currKey = extract(atom) ;
    flg = compare(currKey, key) ;
    if (flg == 0) {
      return this->indElem(curr) + 
	this->tab(curr)->ref()->indKey(key, extract, compare) ;
    } else if (flg == 1) {
      sup = curr ;
    } else {
      inf = curr ;
    }
  }

  tmp = this->max(inf) ;
  atom = &tmp ;
  currKey = extract(atom) ;
  flg = compare(currKey, key) ;
  if (flg != 2) {
    return this->indElem(inf) + 
      this->tab(inf)->ref()->indKey(key, extract, compare) ; 
  } else {
    tmp = this->max(sup) ;
    atom = &tmp ;
    currKey = extract(atom) ;
    flg = compare(currKey, key) ;
    if (flg != 2) {
      return this->indElem(sup) + 
	this->tab(sup)->ref()->indKey(key, extract, compare) ;
    } else {
      return COL_TAIL ;
    }
  } 
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictNode::indElem(int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int currInd = 0 ;
  int ptr = 0 ;
  while (ind > 0) {
    currInd += ((ColNodeElem*)this->tab(ptr))->card ;
    ind-- ; ptr++ ;
  }
  return currInd ; 
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictNode::yoo_refresh(void* addr)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  YOO_LOAD_TIME = TRUE ;
  new(addr) DictNode ;
  YOO_LOAD_TIME = FALSE ;
}


